"""
CRUDE stands for CRUD-Execution.
It is a method to solve the problem of dealing with complex python objects in an
environment that doesn't natively support these.

The method's trick is to allow the complex object's that we "crudified" to be controlled
via a string key that references the complex object, via a "store" which maps
these string keys to the actual physical object.
This store could be a python dictionary (so in RAM) or any persisting storage system
(files, DB) that is given a `typing.Mapping` interface
(see https://i2mint.github.io/dol/ or https://i2mint.github.io/py2store for
tools to do so).

Take, for instance, a GUI that allows a user to compute some descriptive statistics
of the columns of a table.
The inputs are a table, and one of the following statistics function:
``statistics.mean``, ``statistics.median``, or ``statistics.stdev``.

Python functions are not a type natively handled by GUI, so what can we do?
We can stick a layer between our ``compute_stats(stats_func, table)`` function
and our GUI, endowed with a
``{"mean": statistics.mean``, "median": statistics.median, "std": statistics.stdev}``
mapping. We expose the string keys to the GUI, and map them to the functions before
calling ``compute_stats``.

In the case of the ``table``, we'd probably add a means for the GUI user to upload
tables (say from ``.csv`` or ``.xlsx`` files), storing them under a name of their
choice, then pointing to said table via the name, when they want to execute a
``compute_stats(stats_func, table)``.

These are examples of what we call "crudifying" variables or functions.

Here we therefore offer tools to do this sort of thing;
wrap functions so that the complex arguments can be specified through a string key
that points to the actual python object (which is stored in a session's memory or
persisted in some fashion).
"""

from typing import Any, Mapping, Optional, Callable, Union, Iterable, Iterator
from inspect import Parameter
import os
from functools import partial

import dill  # pip install dill

from i2 import Sig, double_up_as_factory
from i2.wrapper import Ingress, wrap
from dol import Files, wrap_kvs
from dol.filesys import mk_tmp_dol_dir, ensure_dir


KT = str
VT = Any
StoreType = Mapping[KT, VT]
StoreName = KT
Mall = Mapping[StoreName, StoreType]


def auto_key(*args, **kwargs) -> KT:
    """Make a str key from arguments.

    >>> auto_key(1,2,c=3,d=4)
    '1,2,c=3,d=4'
    >>> auto_key(1,2)
    '1,2'
    >>> auto_key(c=3,d=4)
    'c=3,d=4'
    >>> auto_key()
    ''
    """
    args_str = ','.join(map(str, args))
    kwargs_str = ','.join(map(lambda kv: f'{kv[0]}={kv[1]}', kwargs.items()))
    return ','.join(filter(None, [args_str, kwargs_str]))


@wrap_kvs(data_of_obj=dill.dumps, obj_of_data=dill.loads)
class DillFiles(Files):
    """Serializes and deserializes with dill"""

    pass


def mk_mall_of_dill_stores(store_names=Iterable[StoreName], rootdir=None):
    """Make a mall of DillFiles stores"""
    rootdir = rootdir or mk_tmp_dol_dir('crude')
    if isinstance(store_names, str):
        store_names = store_names.split()

    def name_and_rootdir():
        for name in store_names:
            root = os.path.join(rootdir, name)
            ensure_dir(root)
            yield name, root

    return {name: DillFiles(root) for name, root in name_and_rootdir()}


from functools import wraps
from i2 import call_forgivingly


def _validate_function_keyword_only_params(func, allowed_params: Iterable, obj_name):
    if func is not None:
        if not callable(func):
            raise TypeError(f'{obj_name} should be callable: {func}')
        sig = Sig(func)
        if not all(kind == Parameter.KEYWORD_ONLY for kind in sig.kinds.values()):
            raise TypeError(f'All params of {obj_name} must be keyword-only. {sig}')
        if not set(sig.names).issubset(allowed_params):
            raise TypeError(f'All params of {obj_name} must be in {allowed_params}')


def _store_on_output(*args, _store_on_ouput_args, **kwargs):
    # TODO: Assuming order is dangerous. Add extra safety. E.g. extract from dict.
    (
        func,
        sig,
        store,
        save_name_param,
        empty_name_callback,
        auto_namer,
        output_trans,
    ) = _store_on_ouput_args
    arguments = sig.kwargs_from_args_and_kwargs(args, kwargs, apply_defaults=True)
    save_name = arguments.pop(save_name_param)
    if not save_name and empty_name_callback:
        assert callable(
            empty_name_callback
        ), f'empty_name_callback must be callable: {empty_name_callback}'
        empty_name_callback()
    args, kwargs = sig.args_and_kwargs_from_kwargs(arguments)
    output = func(*args, **kwargs)
    if not save_name and auto_namer:
        save_name = call_forgivingly(auto_namer, arguments=arguments, output=output)
    if save_name:
        store[save_name] = output
    if not output_trans:
        return output
    else:
        return call_forgivingly(
            output_trans, save_name=save_name, output=output, arguments=arguments
        )


# TODO: store_on_output not pickalable: extend i2.wrapper to be able to solve with it
@double_up_as_factory
def store_on_output(
    func=None,
    *,
    store=None,
    save_name_param='save_name',
    add_store_to_func_attr='output_store',
    empty_name_callback: Callable[[], Any] = None,
    auto_namer: Callable[..., str] = None,
    output_trans: Callable[..., Any] = None,
):
    """Wrap func so it will have an extra save_name_param that can be used to
    indicate whether to save the output of the function call to that key, in
    that store.

    The store can be specified, but an empty dict will be made for it by default.

    By default, a pointer to this store is added to the wrapped function so it can be
    used for inspection etc.

    :param func: Function that we're wrapping
    :param store: Mapping (e.g. dict) where we'd like the output to be stored in
    :param save_name_param: Name of the extra param that will be added to the
        function's interface, where the user will be able to specify the key where
        they want to save the output.
    :param add_store_to_func_attr: If not None, the wrapped function will have an
        attribute of that name where the store will be stored
    :param empty_name_callback: If not None, will be called if the user doesn't specify
        a save name. Intended use; raising errors to force the user to enter name.
    :param auto_namer: If not None, should be a keyword-only ``(*, arguments, output)``
        callback that will be called (forgivingly). This callback will
        be called when the user doesn't specify a name (or an empty name). It's output
        will be used as the ``save_name``.
    :param output_trans: If not None, should be a keyword-only
        ``(*, save_name, output, arguments)`` function that will be called, returning
        it's result instead of the ``output``.
    :return: A wrapped function that has an extra param (default is ``save_name``) that
        allowing the user to specify a key under which to save the output.

    A simple example:

    >>> def foo(a, b: int = 2):
    ...     return a + b
    >>> f = store_on_output(foo)
    >>> f(2, 3, save_name='take')
    5
    >>> f.output_store
    {'take': 5}

    An example with auto_namer. Note that an auto_namer can have no parameters (e.g.
    return a stringified timestamp for the auto-name), but if it does have any
    parameters, these parameters must be keyword-only and the only names allowed are:

    - ``save_name``, which will contain the value of the name the user entered

    - ``output``, which will contain the return value of the function call

    - ``arguments``, which will contain a dict of all the arguments of the function call

    In the following, we will make an ``output_trans`` that just returns the
    ``save_name``. This is useful when you want to make a pipeline and tell the next
    function where to find the output of a previous function.

    >>> def return_key(*, save_name):
    ...     return save_name
    >>> g = store_on_output(foo, output_trans=return_key)
    >>> g(100, 23, save_name='test')
    'test'
    >>> g.output_store
    {'test': 123}
    >>> g.output_store[g(40, 2, save_name='here')]
    42

    An example involving more params:

    >>> my_store = {'all': 'mine'}
    >>> @store_on_output(
    ...     store=my_store,
    ...     save_name_param='save_as',
    ...     add_store_to_func_attr=None,
    ...     auto_namer=lambda *, arguments: '_'.join(map(str, arguments.values()))
    ... )
    ... def bar(a, b: int = 2):
    ...     return a + b
    >>> bar(2, 3, save_as='test')
    5
    >>> my_store
    {'all': 'mine', 'test': 5}
    >>> bar(7)
    9
    >>> my_store
    {'all': 'mine', 'test': 5, '7_2': 9}

    """
    save_name_param_obj = Parameter(
        name=save_name_param, kind=Parameter.KEYWORD_ONLY, default='', annotation=str,
    )
    _validate_function_keyword_only_params(
        auto_namer, ['output', 'arguments'], obj_name='auto_namer'
    )
    _validate_function_keyword_only_params(
        output_trans, ['save_name', 'output', 'arguments'], obj_name='output_trans'
    )
    if output_trans:
        assert callable(output_trans) and set(Sig(output_trans).names).issubset(
            ['save_name', 'output', 'arguments']
        )
    sig = Sig(func) + [save_name_param_obj]

    if store is None:
        store = dict()

    __func = partial(
        _store_on_output,
        _store_on_ouput_args=(
            func,
            sig,
            store,
            save_name_param,
            empty_name_callback,
            auto_namer,
            output_trans,
        ),
    )

    new_return_annotation = (
        Sig(output_trans).return_annotation if output_trans else Parameter.empty
    )
    if new_return_annotation != Parameter.empty:
        sig = sig.replace(return_annotation=new_return_annotation)

    _func = sig(wraps(func)(__func))

    if isinstance(add_store_to_func_attr, str):
        setattr(_func, add_store_to_func_attr, store)

    return _func


@double_up_as_factory
def prepare_for_crude_dispatch(
    func: Callable = None,
    *,
    param_to_mall_map: Optional[Union[dict, Iterable]] = None,
    mall: Optional[Mall] = None,
    include_stores_attribute: bool = False,
    output_store: Optional[Union[Mapping, str]] = None,
    # the arguments below only apply if output_store is given
    save_name_param: str = 'save_name',
    empty_name_callback: Callable[[], Any] = None,
    auto_namer: Callable[..., str] = None,
    output_trans: Callable[..., Any] = None,
):
    """
    Wrap func into something that is ready for CRUDE dispatch.
    It will be a function for whom specific arguments will be specified by strings,
    via underlying stores containing the values.
    We say that those arguments were crude-dispatched.

    :param func: callable, the function to wrap
    :param param_to_mall_map: dict, whose keys specify which params should be
        crude-dispatched and whose values are the mall keys to the Mapping instances
        (e.g. dict) that should be used for said param.
        If a non-Mapping iterable is given, will take {name: name...} identity mapping
        for names in that iterable.
    :param mall: A store of stores. A Mapping whose keys are what the values of
        ``param_to_mall_map`` point to and whose values are mapping interfaces (called
         "stores" of a storage backend (local or remote, persisted or in-memory).
    :param include_stores_attribute: bool, whether to add an attribute to the function
        containing the ``output_store``
    :param output_store: a store used to record the output of the function

    If (and only if) ``output_store`` is given, we can also use the following
    arguments to control the output "crudification" further.

    :param save_name_param: str, the argument name that should be used in the returned
        functions to get the the key of ``output_store`` under which the output will be
        saved.
    :param empty_name_callback: If not None, will be called when the user (of the
        wrapped function ``prepare_for_crude_dispatch`` will output) doesn't
        specify a save name. Intended use; raising errors (or other kinds of behavior)
        when we want to force the user to enter name.
    :param auto_namer: If not None, should be a keyword-only ``(*, arguments, output)``
        callback that will be called (forgivingly). This callback will
        be called when the user doesn't specify a name (or an empty name). It's output
        will be used as the ``save_name``. Intended use is to produce a function where
        that has auto-naming capabilities.
    :param output_trans: If not None, should be a keyword-only
        ``(*, save_name, output, arguments)`` function that will be called, returning
        it's result instead of the ``output``.

    :return: A function that outputs the same thing as ``func``, but (1) with some
        parameters being changed so that on can specify some arguments
        (those listed by ``param_to_mall_key_dict``)


    >>> def func(a, b: float, c: int):
    ...     return a + b * c
    ...
    >>> param_to_mall_map = dict(a='a', b='b_store')
    >>>
    >>> mall = dict(
    ...     a = {'one': 1, 'two': 2},
    ...     b_store = {'three': 3, 'four': 4},
    ...     unused_store = {'to': 'illustrate'}
    ... )
    >>> crude_func = prepare_for_crude_dispatch(
    ...     func, param_to_mall_map=param_to_mall_map, mall=mall
    ... )

    ``crude_func`` is like ``func``, but you enter your ``a`` and ``b`` inputs
    via string keys that will look up the values you're pointing to in ``mall``

    >>> assert crude_func('one', 'three', 10) == func(1, 3, 10) == 31
    >>> crude_func('one', 'three', 10)
    31
    >>> func(1, 3, 10)
    31

    What's happening behind the scenes of ``crude_func`` is this.
    Let's follow what happens for the ``b`` argument. Say the user says ``b='three'``...

    - ``b`` is a key of ``param_to_mall_key_dict``, indicating that it should be "cruded"

    - ``param_to_mall_key_dict['b']`` is ``'b_store'``, so we know which mall key to use

    - We retrieve ``mall['b_store']['three']``, which is ``3``, and call ``func`` with it

    So this is equivalent to this:

    >>> func(mall['a']['one'], mall['b_store']['three'], 10)
    31

    The signature of ``a`` and ``b`` also changed to be `str`:

    >>> from inspect import signature
    >>> str(signature(crude_func))
    '(a: str, b: str, c: int)'

    By default, the ``output_store`` argument is None, but if you specify a mapping
    there (or a string key that appears in the ``mall`` you specified, pointing to
    a mapping), then the function you'll get will have an extra argument.

    >>> output_store = dict()
    >>> crude_func = prepare_for_crude_dispatch(
    ...     func,
    ...     param_to_mall_map=param_to_mall_map,
    ...     mall=mall,
    ...     output_store=output_store
    ... )
    >>> str(signature(crude_func))
    "(a: str, b: str, c: int, save_name: str = '')"

    You now have this extra ``save_name`` param in your function.
    (Note that you can change its name through the ``prepare_for_crude_dispatch``'s
    ``save_name_param`` argument.)
    The default for ``save_name`` is '', and if you don't specify a non-empty
    string, nothing different will happen, but if you do specify a non-empty string,
    the output of your function will be saved, in the ``output_store`` you specified,
    using that ``save_name`` key you specified.

    >>> crude_func('one', 'three', 10)
    31
    >>> output_store
    {}
    >>> crude_func('one', 'three', 10, save_name='save_here')
    31
    >>> output_store
    {'save_here': 31}

    The example below shows how one can

    - change the default ``save_name_param``

    - use an ``auto_namer``

    - use ``output_trans`` to return both save name and output

    >>> def bar(a, b: int = 2):
    ...     return a + b
    >>> my_store = {'all': 'mine'}
    >>> def return_key_and_val(*, save_name, output):
    ...     return save_name, output
    >>> wbar = prepare_for_crude_dispatch(
    ...     bar,
    ...     output_store=my_store,
    ...     save_name_param='save_as',
    ...     auto_namer=lambda *, arguments: '_'.join(map(str, arguments.values())),
    ...     output_trans=return_key_and_val
    ... )

    >>> wbar(2, 3, save_as='test')  # output will be the (save_as, output) pair
    ('test', 5)
    >>> my_store  # but only the output is saved (under save_as key)
    {'all': 'mine', 'test': 5}
    >>> wbar(7)  # and if you don't specify a save_as key, one is made for you!
    ('7_2', 9)
    >>> my_store
    {'all': 'mine', 'test': 5, '7_2': 9}
    """
    ingress = None

    store_for_param = {}

    if param_to_mall_map is not None:

        sig = Sig(func)

        # get an {argname: store, ...} dict from param_to_mall_map:
        store_for_param = _mk_store_for_param(sig, param_to_mall_map, mall) or dict()

        def kwargs_trans(outer_kw):
            """
            Let's say you have a function with three params: a, b, and c, whose arguments
            should be ints. Let's say you want a and c to be cruded.
            Then you need to specify a store for each one of these:

            >>> store_for_param = {
            ...     'a': {'one': 1, 'two': 2},
            ...     'c': {'three': 3}
            ... }

            What kwargs_trans will with this store_for_param, is this:

            >>> kwargs_trans({'a': 'one', 'b': 2, 'c': 'three'})
            {'a': 1, 'b': 2, 'c': 3}
            """
            # outer_kw is going to be the new/wrapped/cruded interface of the function
            # That is, the one that takes strings to specify arguments
            # What we need to do now is transform the cruded argument values from strings
            # to the values these strings are pointing to (via the store corresponding
            # to that argument).
            def get_values_from_stores():
                # Note: only need to specify arguments that change
                for param, store in store_for_param.items():
                    # param's argument value is assumed to be a store_key
                    store_key = outer_kw[param]
                    # store_key points to the value the outer user wants the value for:
                    # store is the store where to find it
                    yield param, store[store_key]

            return dict(get_values_from_stores())

        ingress = Ingress(
            inner_sig=sig,
            kwargs_trans=kwargs_trans,
            outer_sig=(
                sig.ch_annotations(**{name: str for name in store_for_param})
                # + [save_name_param]
            ),
        )

    wrapped_f = wrap(func, ingress=ingress)

    if include_stores_attribute:
        wrapped_f.store_for_param = store_for_param

    if output_store is not None:
        output_store_name = 'output_store'
        if isinstance(output_store, str):
            # if output_store is a string, it should be the a key to store_for_param
            output_store_name = output_store
            output_store = mall[output_store_name]
        else:
            # TODO: Assert MutableMapping, or just existence of __setitem__?
            if not hasattr(output_store, '__setitem__'):
                raise ValueError(f'Needs to have a __setitem__: {output_store}')
        if output_store_name in store_for_param:
            raise ValueError(
                f'Name conflicts with existing param name: {output_store_name}'
            )

        wrapped_f = store_on_output(
            wrapped_f,
            store=output_store,
            save_name_param=save_name_param,
            add_store_to_func_attr='output_store' if include_stores_attribute else None,
            empty_name_callback=empty_name_callback,
            auto_namer=auto_namer,
            output_trans=output_trans,
        )
        # wrapped_f.__name__ = wrapped_f.__name__ + "_w_output_storing"  # to remove

    return wrapped_f


def _mk_store_for_param(sig, param_to_mall_key_dict=None, mall=None, verbose=True):
    """Make a {param: store,...} dict from a {param: mall_key,...} dict, a sig and a
    mall, validating stuff on the way."""
    mall = mall or dict()
    param_to_mall_key_dict = keys_to_values_if_non_mapping_iterable(
        param_to_mall_key_dict
    )
    # mall_keys_that_are_also_params_but_not_param_to_mall_key_dict_keys
    unmentioned_mall_keys = set(mall) & set(sig.names) - set(param_to_mall_key_dict)
    if unmentioned_mall_keys and verbose:
        from warnings import warn

        warn(
            f"Some of your mall keys were also func arg names, but you didn't mention "
            f'them in param_to_mall_map, namely, these: {unmentioned_mall_keys}'
        )
    if param_to_mall_key_dict:
        func_name_stub = ''
        if sig.name:
            func_name_stub = f'({sig.name})'
        if isinstance(param_to_mall_key_dict, str):
            param_to_mall_key_dict = param_to_mall_key_dict.split()
        if not set(param_to_mall_key_dict).issubset(sig.names):
            offenders = set(param_to_mall_key_dict) - set(sig.names)
            raise ValueError(
                'The param_to_mall_map should only contain keys that are '
                f"parameters (i.e. 'argument names') of your function {func_name_stub}. "
                f'Yet these param_to_mall_map keys were not argument names: '
                f'{offenders}'
            )
        if not set(param_to_mall_key_dict.values()).issubset(mall.keys()):
            offenders = set(param_to_mall_key_dict.values()) - set(mall.keys())
            keys = 'keys' if len(offenders) > 1 else 'key'
            offenders = ', '.join(map(lambda x: f"'{x}'", offenders))

            raise ValueError(
                f'The {offenders} {keys} of your param_to_mall_map values were not '
                f'in the mall. Your param_to_mall_key_dict is:\n'
                f'{param_to_mall_key_dict} and your mall has keys: {list(mall)}. '
                f'You can either add {offenders} stores to the mall, '
                # f'make an auto-store-making mall (e.g. collections.defaultdict(dict), '
                f'or change your param_to_mall_map to point (values) to a store that '
                f'you actually have in the mall'
            )
        # Note: store_for_param used to be the argument of prepare_for_crude_dispatch,
        #   instead of the (param_to_mall_map, mall) pair which is overkill.
        #   The reason for obliging the user to give this pair was because asking for
        #   the user to be more explicit about the argname to store mapping would avoid
        #   some bugs and make it possible to validate the request earlier on.
        store_for_param = {
            argname: mall[mall_key]
            for argname, mall_key in param_to_mall_key_dict.items()
        }
        return store_for_param


def keys_to_values_if_non_mapping_iterable(d: Optional[Iterable]) -> dict:
    if d is None:
        return dict()
    elif not isinstance(d, Mapping) and isinstance(d, Iterable):
        d = {k: k for k in d}
    return d


# Note: This is not meant to actually be used in real apps, but be a drop in helper to
#   talk to the mall (or rather "listen" since it's read-only) from a UI.
def simple_mall_dispatch_core_func(
    key: KT, action: str, store_name: StoreName, mall: Mall
):
    """Helper function to dispatch a mall

    This function is only meant to be a helper to give a UI (GUI,
    CLI...) mall-exploration capabilities. Namely:

    - ``list(mall)``: list the keys of a mall. This is achieved with args:
        ``(key=None, action=None, store_name=None, mall=mall)``
    - ``mall[store_name]``: get a store. Acheived by:
        ``(key=None, action=None, store_name=store_name, mall=mall)``
    - ``list(mall[store_name])``: list keys of a store (of the mall). Acheived by:
        ``(key=None, action='list', store_name=store_name, mall=mall)``
    - ``list(filter(key, mall[store_name]))``: list keys of a store (of the mall)
        according to a substring filter. (only keys that have ``key`` as substring)
        ``(key=key, action='list', store_name=store_name, mall=mall)``
    - ``mall[store_name][key]``:  get the value/data of a store for ``key``
        ``(key=key, action='get', store_name=store_name, mall=mall)``

    :param key: The key
    :param action: 'list' (to list keys of a store) or 'get' (to get the value of
        ``key`` in the store (named ``store_name``)
    :param store_name: Store name to look up in mall. If not given, the function will
        output the mall keys (which are valid store names)
    :param mall: dict of stores (Mapping interface to data)
    :return:

    >>> mall = {
    ...     'english': {'one': 1, 'two': 2, 'three': 3},
    ...     'french': {'un': 1, 'deux': 2},
    ... }

    List the keys of a mall:

    >>> simple_mall_dispatch_core_func(None, None, None, mall=mall)
    ['english', 'french']

    Get a store

    >>> simple_mall_dispatch_core_func(None, None, store_name='english', mall=mall)
    {'one': 1, 'two': 2, 'three': 3}

    List keys of a store (of the mall):

    >>> simple_mall_dispatch_core_func(
    ...     None, action='list', store_name='english', mall=mall
    ... )
    ['one', 'two', 'three']

    List keys of a store (of the mall) according to a substring filter:

    >>> simple_mall_dispatch_core_func(
    ...     'e', action='list', store_name='english', mall=mall
    ... )
    ['one', 'three']

    >>> simple_mall_dispatch_core_func(
    ...     'two', action='get', store_name='english', mall=mall
    ... )
    2


    """
    if not store_name:
        # if store_name empty, list the store names (i.e. the mall keys)
        return list(mall)
    else:  # if not, get the store
        store = mall[store_name]
        if not action:
            return store

    key = key or ''
    if action == 'list':
        key = key.strip()  # to handle some invisible whitespace that would screw things
        return list(filter(lambda k: key in k, store))
    elif action == 'get':
        return store[key]



from typing import Callable
from i2 import Sig


def implement_component(
    base_cls: type,
    component_factory: Callable,
    input_value_callback: Callable = None,
    **input_mapping
):
    component_factory_sig = Sig(component_factory)
    keyword_names = component_factory_sig.keyword_names

    class component_class(base_cls):
        def render(self):
            component_factory_kwargs = {}
            for name in keyword_names:
                attr_name = input_mapping.get(name, name)
                attr = getattr(self, attr_name, None)
                if attr is not None:
                    component_factory_kwargs[name] = attr
            input_value = component_factory(**component_factory_kwargs)
            if input_value_callback:
                input_value_callback(input_value, self)
            return input_value

    return component_class



import os


def foo(text: str, integer: int, floating: float = 1.0):
    """C'est foo"""
    return text, integer, floating


def bar(a: str = 'boo'):
    """About bar"""
    return a


funcs = [foo, bar]

from streamlitfront.base import (
    _get_state,
    _get_view_key,
    get_pages_specs,
    dflt_hash_funcs,
)

import streamlit as st


def pages_app(funcs, configs):
    state = _get_state(hash_funcs=dflt_hash_funcs)  # TODO: get from configs

    # Page setup
    st.set_page_config(layout='wide')

    # Make page objects
    views = get_pages_specs(funcs, **configs)
    state['views'] = views

    # TODO: The above is static: Should the above be done only once, and cached?
    #   Perhaps views should be cached in state?

    # Setup navigation
    view_key = _get_view_key(tuple(views.keys()), label='Select your view')

    # Display the selected page with the session state
    # This is the part that actually runs the functionality that pages specifies
    view_runner = views[view_key]  # gets the page runner
    view_runner(state)  # runs the page with the state

    # Mandatory to avoid rollbacks with widgets, must be called at the end of your app
    state.sync()


if __name__ == '__main__':
    from streamlitfront.base import dispatch_funcs
    from streamlitfront.page_funcs import ExperimentalViewFunc

    print('file: {}'.format(os.path.realpath(__file__)))

    app = dispatch_funcs(funcs, configs=dict(page_factory=ExperimentalViewFunc))

    app()



import os


def foo(state, text: str, options: list = []):
    return text, options


def bar(state, a: str = 'boo'):
    return a


funcs = [foo, bar]

if __name__ == '__main__':
    from streamlitfront.base import dispatch_funcs
    from streamlitfront.page_funcs import DataBindingExploPageFunc

    print('file: {}'.format(os.path.realpath(__file__)))

    app = dispatch_funcs(funcs, configs=dict(page_factory=DataBindingExploPageFunc))

    app()



from abc import ABC
from typing import Any, Callable, Iterable
from front.elements import ElementTreeMaker
from front.elements.elements import FrontContainerBase
from front.spec_maker_base import SpecMakerBase
from front.types import FrontApp, Map


class AppMaker:
    def __init__(
        self,
        spec_maker_factory: Callable,
        element_tree_maker_factory: Callable = ElementTreeMaker,
    ):
        self.spec_maker: SpecMakerBase = spec_maker_factory()
        self.el_tree_maker: ElementTreeMaker = element_tree_maker_factory()

    def mk_app(
        self, objs: Iterable[Any], config: Map = None, convention: Map = None
    ) -> FrontApp:
        element_tree, app_spec = self._element_tree_and_spec(objs, config, convention)
        return self._mk_app(element_tree, app_spec)

    # TODO: Find a way so this method does not do two things. Maybe consider the
    # element tree to be part of the spec object.
    def _element_tree_and_spec(
        self, objs: Iterable[Any], config: Map = None, convention: Map = None
    ):
        spec = self.spec_maker.mk_spec(config, convention)
        front_objs = self._prepare_objs(objs, spec.obj_spec)
        element_tree = self.el_tree_maker.mk_tree(front_objs, spec.rendering_spec)
        return element_tree, spec.app_spec

    def _prepare_objs(self, objs: Iterable[Any], obj_spec: dict) -> Iterable[Any]:
        def validate_obj(obj):
            if not isinstance(obj, Callable):
                raise NotImplementedError(
                    'Only objects of type Callable are supported for now.'
                )

        for obj in objs:
            validate_obj(obj)

        trans_func = obj_spec['trans']
        return trans_func(objs)

    def _mk_app(self, element_tree: FrontContainerBase, app_specs: dict) -> FrontApp:
        element_tree.name = app_specs['title']
        return element_tree.render



"""How do we eliminate some of the boilerplate between having python functions
and making them pydantic?

>>> from i2.tests.objects_for_testing import formula1
>>> pyd_input_model = func_to_pyd_input_model_cls(formula1)
>>> pyd_input_model
<class 'pydantic.main.formula1'>
>>> from i2 import Sig
>>> Sig(formula1)
<Sig (w, /, x: float, y=1, *, z: int = 1)>
>>> Sig(pyd_input_model)
<Sig (*, w: Any, x: float, z: int = 1, y: int = 1) -> None>

>>> pyd_func = func_to_pyd_func(formula1)
>>> input_model_instance = pyd_input_model(w=1, x=2)
>>> input_model_instance
formula1(w=1, x=2.0, z=1, y=1)
>>> pyd_func(input_model_instance)
3.0
>>> formula1(1, x=2)  # can't say w=1 because w is position only
3

"""

from typing import Callable, Any
from functools import partial

from pydantic import create_model, BaseModel

from i2 import Sig, name_of_obj, empty_param_attr
from i2.wrapper import wrap, Ingress


def pyd_func_ingress_template(input_model_instance, wrapped_func_sig: Sig):
    kwargs = dict(input_model_instance)
    args, kwargs = wrapped_func_sig.args_and_kwargs_from_kwargs(kwargs)
    return args, kwargs


# TODO: Add the output model annotation
def func_to_pyd_func(func: Callable, dflt_type=Any):
    """Get a 'opyrator' function from a python function.
    That is, a function that has a single pydantic model input and output.
    """
    pyd_func_ingress = partial(pyd_func_ingress_template, wrapped_func_sig=Sig(func))

    input_model = func_to_pyd_input_model_cls(func, dflt_type)
    output_model = create_model(
        'output_model', output_val=(Any, ...)
    )  # TODO: Work on this
    # TODO: Inject annotations in pyd_func_ingress

    return wrap(func, ingress=pyd_func_ingress)


def func_to_pyd_input_model_cls(
    func: Callable, dflt_type=Any, *, name=None, warn_when_changing_names=True
):
    """Get a pydantic model of the arguments of a python function

    >>> def foo(a, b: int, c: bool=False):
    ...     ...
    >>> obj = func_to_pyd_input_model_cls(foo)
    >>> import json
    >>> assert json.loads(obj.schema_json()) == (
    ... {
    ...     'title': 'foo',
    ...     'type': 'object',
    ...     'properties': {
    ...         'a': {'title': 'A'},
    ...         'b': {'title': 'B', 'type': 'integer'},
    ...         'c': {'title': 'C', 'default': False, 'type': 'boolean'}
    ...     },
    ...     'required': ['a', 'b']
    ... })

    If some argument names of the function conflict with attribute names of BaseModel,
    these will be capitalized to resolve the conflict.

    >>> def bar(x, copy, schema):
    ...     ...
    >>> obj2 = func_to_pyd_input_model_cls(bar, warn_when_changing_names=False)
    >>> assert json.loads(obj2.schema_json()) == (
    ... {
    ...     'title': 'bar',
    ...     'type': 'object',
    ...     'properties': {
    ...         'x': {'title': 'X'},
    ...         'COPY': {'title': 'Copy'},
    ...         'SCHEMA': {'title': 'Schema'}},
    ...     'required': ['x', 'COPY', 'SCHEMA']
    ... })

    """
    name = name or name_of_obj(func)
    try:
        return create_model(name, **dict(func_to_pyd_model_specs(func, dflt_type)))
    except NameError:
        conflicting_names = set(Sig(func).names) & set(dir(BaseModel))
        old_to_new_names = {k: k.upper() for k in conflicting_names}
        if warn_when_changing_names:
            from warnings import warn

            warn(
                f'''{len(conflicting_names)} argument name(s) conflicted with BaseModel.
            They're being replaced with upper-case names to resolve conflict. old:new ->
            {old_to_new_names}
            '''
            )
        wrapped_func = Ingress.name_map(func, **old_to_new_names).wrap(func)
        return create_model(
            name, **dict(func_to_pyd_model_specs(wrapped_func, dflt_type))
        )


def func_to_pyd_model_specs(func: Callable, dflt_type=Any):
    """Helper function to get field info from python signature parameters"""
    for p in Sig(func).params:
        if p.annotation is not empty_param_attr:
            if p.default is not empty_param_attr:
                yield p.name, (p.annotation, p.default)
            else:
                yield p.name, (p.annotation, ...)
        else:  # no annotations
            if p.default is not empty_param_attr:
                yield p.name, p.default
            else:
                yield p.name, (dflt_type, ...)


def pydantic_egress(output):
    return_type = type(output)
    mod = create_model('Output', output_val=return_type)

    return mod(output_val=output)


def pydantic_model_from_type(mytype, name='Output', field_name='result'):
    model = create_model(name, **{field_name: (mytype, ...)})

    return model



"""Base functions for front dispatching"""

from functools import partial
from typing import Mapping, Optional
from i2 import double_up_as_factory

from front.util import inject_enum_annotations, subdict
from front.crude import (
    keys_to_values_if_non_mapping_iterable,
    prepare_for_crude_dispatch,
)


# Note: Could also use a Pipe to make the dispatcher (more composable?)
# Note: Should evolve to a parametrized pipeline of preparers.
@double_up_as_factory
def prepare_for_dispatch(
    func=None,
    *,
    # for prepare_for_crude_dispatch
    param_to_mall_map=(),
    mall=None,
    output_store=None,
    save_name_param: str = 'save_name',
    include_stores_attribute: bool = False,
    # for setting defaults
    defaults: Optional[Mapping] = None,
):
    param_to_mall_map = keys_to_values_if_non_mapping_iterable(param_to_mall_map)

    from i2 import Sig

    wrapped_func = func  # to seed the following sequence of wrapping

    # crude wrap
    wrapped_func = prepare_for_crude_dispatch(
        wrapped_func,
        param_to_mall_map=param_to_mall_map,
        mall=mall,
        output_store=output_store,
        save_name_param=save_name_param,
        include_stores_attribute=include_stores_attribute,
    )

    # enum wrap
    wrapped_func = inject_enum_annotations(
        wrapped_func,
        extract_enum_value=True,
        **{param: mall[mall_key] for param, mall_key in param_to_mall_map.items()},
    )

    # insert defaults
    if defaults:
        # get only keys that are in wrapped_func signature as well as defaults
        wrapped_sig = Sig(wrapped_func)
        _defaults = wrapped_sig.kwargs_from_args_and_kwargs(
            (), defaults or {}, allow_partial=True, allow_excess=True
        )
        _defaults_args, _defaults_kwargs = wrapped_sig.args_and_kwargs_from_kwargs(
            _defaults, allow_partial=True, allow_excess=True
        )
        wrapped_func = partial(wrapped_func, *_defaults_args, **_defaults_kwargs)

    return wrapped_func



"""Dispatching python functions as webservices, docker containers, and GUIs

Consider these three functions:

>>> def foo(a: int = 0, b: int = 0, c=0):
...     'This is foo. It computes something'
...     return (a * b) + c
>>> def bar(x, greeting='hello'):
...     'bar greets its input'
...     return f'{greeting} {x}'
>>> def confuser(a: int = 0, x: float = 3.14):
...     return (a ** 2) * x

The objective here is to be able to do this:

>>> app = dispatch_funcs([foo, bar, confuser], ...)  # doctest: +SKIP

getting a deployable app that allows the user to operate with these three wonderful
functions. The ellipses (`...`) are there to indicate that we may want to specify
the kind of app we want (web-service, GUI, CLI...) as well as particular configurations
for the latter.

"""


from contextlib import suppress

# TODO: Find a better way to do this (plugin) thing
with suppress(ModuleNotFoundError, ImportError):
    from streamlitfront import page_funcs
    from streamlitfront import session_state
    from streamlitfront import base
    from streamlitfront import util
    from streamlitfront import *
    from warnings import warn

    warn('Moved to seperate streamlitfront package', DeprecationWarning)

from .app_maker import AppMaker
from .spec_maker_base import (
    SpecMakerBase,
    APP_KEY,
    OBJ_KEY,
    RENDERING_KEY,
    NAME_KEY,
)
from .elements import (
    ElementTreeMaker,
    ELEMENT_KEY,
    FrontComponentBase,
    FrontContainerBase,
    FrontElementBase,
    DEFAULT_INPUT_KEY,
)



from functools import partial
from random import choice, randint, uniform
import string
from i2 import Sig


rdm_int = partial(randint, a=-100, b=100)
rdm_float = partial(uniform, a=-100.0, b=100.0)


def rdm_str():
    nb_char = randint(5, 15)
    return ''.join(choice(string.ascii_letters + string.digits) for _ in range(nb_char))


PRIMITIVE_TYPES = [bool, str, int, float]


def get_var_nodes_to_crudify(funcs):
    var_nodes = set()
    for func in funcs:
        sig = Sig(func)
        for param, annotation in sig.annotations.items():
            if annotation not in PRIMITIVE_TYPES:
                var_nodes.add(param)
    return list(var_nodes)



from front.py2pydantic import *

# ---------------------------------------------------------------------------------------
# Tests

# This one has every of the 4 combinations of (default y/n, annotated y/n)
# def formula1(w, /, x: float, y=1, *, z: int = 1):
def formula1(w, x: float, y=1, z: int = 1):
    return ((w + x) * y) ** z


def test_func_to_pyd_model_of_inputs(func: Callable = formula1, dflt_type=Any):
    pyd_model = func_to_pyd_input_model_cls(func, dflt_type)
    sig = Sig(func)
    # # TODO: Don't want to use hidden __fields__. Other "official" access to this info?
    # the names of the arguments should correspond to names of the model's fields:
    assert sorted(sig.names) == sorted(pyd_model.__fields__)
    # and for each field, name, default, and sometimes annotations/types should match...
    for name, model_field in pyd_model.__fields__.items():
        param = sig.parameters[name]
        expected_default = (
            param.default if param.default is not empty_param_attr else None
        )
        assert model_field.default == expected_default
        if param.annotation is not empty_param_attr:
            # if arg is annotated, expect that as the field type
            assert model_field.type_ == param.annotation
        elif param.default is empty_param_attr:
            # if arg is not annotated and doesn't have a default, expect dflt_type
            assert model_field.type_ == dflt_type
        # but don't test pydantic's resolution of types from default values



import io
import os
from py2store import ZipReader, filt_iter
import pandas as pd
import streamlit as st
from streamlitfront.session_state import _get_state

# store = state_dict[store_name]
# option = st.selectbox(message, [extra] + list(store.keys()), index=dflt_index)


class C:
    def __init__(self):
        self.val = 1

    def increment(self):
        st.write(self.val)
        self.val += 1
        return self.val


c = C()
my_incrementer = c.increment


def display_selectbox(store):
    _keys = filt_iter(store, filt=lambda k: k.endswith('.csv'))
    st.selectbox('Choose a file', list(_keys))


def dacc(zip_path: type(lambda a: a), file: str):  # file: list = []):
    s = ZipReader(zip_path)
    annots_df = pd.read_csv(io.BytesIO(s[file]))
    display_selectbox(s)
    return annots_df


def foo(state: type(_get_state()), word: str):
    return word


funcs = [my_incrementer, foo]

if __name__ == '__main__':
    from streamlitfront.base import dispatch_funcs

    print('file: {}'.format(os.path.realpath(__file__)))

    app = dispatch_funcs(funcs)

    app()



from itertools import chain
from typing import Any, Callable, ContextManager, Iterable, Mapping
from meshed.makers import code_to_dag
from front.dag import crudify_func_nodes

from front.tests.util import get_var_nodes_to_crudify, rdm_int, rdm_str

#####################
## SIMPLE USE CASE ##
#####################


def simple_use_case():
    x = foo(a, b, c)
    y = bar(x, greeting)
    z = confuser(a, x)


def foo_func(a: int = 0, b: int = 0, c=0) -> int:
    """This is foo. It computes something"""
    return (a * b) + c


def bar_func(x: str, greeting='hello') -> str:
    """bar greets its input"""
    return f'{greeting} {x}'


def confuser_func(a: int = 0, x: float = 3.14) -> float:
    return (a ** 2) * x


########################
## RUN MODEL USE CASE ##
########################


def run_model_use_case():
    wfs = get_wfs(src_to_wf, data_src)
    chks = get_chunks(chunker, wfs)
    fvs = get_fvs(featurizer, chks)
    model_outputs = run_model(model, fvs)


def get_wfs_func(src_to_wf: Callable, data_src: int) -> Iterable[Iterable[int]]:
    return src_to_wf(data_src)


def get_chunks_func(
    chunker: Callable, wfs: Iterable[Iterable[int]]
) -> Iterable[Iterable[int]]:
    return chain.from_iterable(map(chunker, wfs))


def get_fvs_func(featurizer: Callable, chks: Iterable[Iterable[int]]):
    return map(featurizer, chks)


def run_model_func(model: Callable, fvs: Iterable[int]) -> Iterable[str]:
    return list(map(model, fvs))


def src_to_wf_func(src):
    for i in range(src):
        yield [i, i + 2, i * 3, i + 5]


def chunker_func(wf):
    yield wf[0:2]
    yield wf[2:4]


featurizer_func = sum


def model_func(x):
    return f'hi {x}'


###########
## TESTS ##
###########

TEST_USE_CASE_PARAMETER_NAMES = 'use_case, func_src, kwargs'
TEST_USE_CASE_PARAMETERS = [
    (
        simple_use_case,
        {'foo': foo_func, 'bar': bar_func, 'confuser': confuser_func,},
        {'a': rdm_int(), 'b': rdm_int(), 'c': rdm_int(), 'greeting': rdm_str(),},
    ),
    # BUG: The execution of a crudified func node through front returns the result
    # of the initial function instead of the key used to store the result.
    # (
    #     run_model_use_case,
    #     {
    #         'get_wfs': get_wfs_func,
    #         'get_chunks': get_chunks_func,
    #         'get_fvs': get_fvs_func,
    #         'run_model': run_model_func,
    #     },
    #     {
    #         'data_src': 5,
    #         'src_to_wf': src_to_wf_func,
    #         'chunker': chunker_func,
    #         'featurizer': featurizer_func,
    #         'model': model_func,
    #     },
    # ),
]


def base_test_use_case(
    use_case: Callable,
    func_src: Mapping[str, Callable],
    inputs: Mapping[str, Any],
    mk_front_func: Callable,
    run_front_app: ContextManager,
):
    def mk_crudified_dag():
        mall = dict()
        for k, v in inputs.items():
            if k in var_nodes_to_crudify:
                store_key = str(v)
                mall[f'{k}_store'] = {store_key: v}
                front_dag_inputs[k] = store_key
        return crudify_func_nodes(var_nodes_to_crudify, dag, mall=mall)

    dispatch_func_src = dict(func_src)
    var_nodes_to_crudify = get_var_nodes_to_crudify(dispatch_func_src.values())
    front_dag_inputs = dict(inputs)
    if var_nodes_to_crudify:
        crudified_dag = mk_crudified_dag()
        dispatch_func_src = {
            func_node.name: func_node.func for func_node in crudified_dag.func_nodes
        }
    dispatch_funcs = list(dispatch_func_src.values())
    with run_front_app(dispatch_funcs) as front_app:
        front_func_src = {
            node_name: mk_front_func(func, dispatch_funcs, front_app)
            for node_name, func in dispatch_func_src.items()
        }
        dag = code_to_dag(use_case, func_src=func_src)
        front_dag = code_to_dag(use_case, func_src=front_func_src)
        assert front_dag(**front_dag_inputs) == dag(**inputs)



import os
import dill
from haggle import KaggleDatasets
import pandas as pd
import numpy as np
import streamlit as st
from slang.snippers import PcaChkToFv, LdaChkToFv
from slang.util import mk_callable

from odat.mdat.local_kgl import mk_dacc
from streamlitfront.session_state import _get_state

# TODO: Replace omodel models with open source ones
from omodel.outliers.outlier_model import OutlierModel as Stroll
from omodel.ml.scent import CentroidSmoothing

# TODO: Resolve the UnhashableTypeError: Cannot hash object of type _thread.Rlock, found in something.


DFLT_FEATURIZER = {'supervised': LdaChkToFv, 'unsupervised': PcaChkToFv}


def search_kaggle(
    keyword_search: str = '', num_to_display: int = 10, metadata: bool = False
):
    s = KaggleDatasets()
    results = s.search(keyword_search)
    if metadata:
        df = pd.DataFrame(results.values())[
            ['ref', 'title', 'subtitle', 'downloadCount', 'totalBytes']
        ]
        df = df.set_index('ref').sort_values('downloadCount', ascending=False)
        st.write(df.head(num_to_display))
    else:
        st.write(list(results)[:10])
    return results


def download_kaggle_dataset(
    state: type(_get_state()), kaggle_path, preview: bool = False
):
    s = KaggleDatasets()
    v = s[kaggle_path]
    if preview:
        st.write(v)
    return kaggle_path


def create_kaggle_dacc(
    # state: type(_get_state()),
    zip_path: type(lambda a: a),
    annots_path: str,
    wrangle_func: type(lambda a: a),
    extension: str = '.wav',
    key_path: str = '',
):
    wrangle_func = dill.load(wrangle_func)
    return mk_dacc(
        zip_path, annots_path, wrangle_func, extension=extension, key_path=key_path,
    )


def build_and_run_model(tag: str, state: type(_get_state())):
    dacc = state['Create Kaggle Dacc']
    chks, tags = zip(*dacc.chk_tag_gen(tag))
    scores, fvs, featurizer, model = produce_results(chks, tags)
    return str(
        'Percentage correct: '
        + str(sum(scores == tags) / len(tags))
        + ' vs. Random guessing: '
        + str(1 / len(set(tags)))
    )


def get_featurizer(chks, tags=None, **kwargs):
    proj_type = 'unsupervised' if (tags is None) else 'unsupervised'
    featurizer = mk_callable('transform')(DFLT_FEATURIZER[proj_type])(**kwargs)
    featurizer.fit(chks, tags)
    return featurizer


def get_outlier_model(fvs, **kwargs):
    # Stroll is already callable, on dim-1 fv
    model = Stroll(**kwargs)
    model.fit(fvs)
    return model


def get_classification_model(fvs, tags, **kwargs):
    CS = mk_callable('predict')(CentroidSmoothing)
    model = CentroidSmoothing(**kwargs)
    return model.fit(fvs, tags)


def produce_results(
    chks,
    tags=None,
    featurizer_kwargs=None,
    classification_model_kwargs=None,
    outlier_model_kwargs=None,
):
    featurizer_kwargs = featurizer_kwargs or {}
    classification_model_kwargs = classification_model_kwargs or {}
    outlier_model_kwargs = outlier_model_kwargs or {}

    featurizer = get_featurizer(chks, tags, **featurizer_kwargs)
    fvs = featurizer(chks)
    if tags is None:
        model = get_outlier_model(fvs, **outlier_model_kwargs)
        scores = np.array(list(map(model, fvs)))
    else:
        model = get_classification_model(fvs, tags, **classification_model_kwargs)
        scores = model(fvs)

    return scores, fvs, featurizer, model


funcs = [
    search_kaggle,
    download_kaggle_dataset,
    create_kaggle_dacc,
    build_and_run_model,
]

if __name__ == '__main__':
    from streamlitfront.base import dispatch_funcs
    from streamlitfront.page_funcs import StatePageFunc

    print('file: {}'.format(os.path.realpath(__file__)))

    app = dispatch_funcs(funcs, configs={'page_factory': StatePageFunc})

    app()



"""Utils"""

from copy import copy
from operator import attrgetter
from functools import partial
from typing import Iterable, Callable, Mapping
from contextlib import suppress
from enum import Enum

from i2 import Sig, double_up_as_factory
from i2.wrapper import Ingress, wrap
from i2.signatures import name_of_obj

from front.types import Map

ignore_import_problems = suppress(ImportError, ModuleNotFoundError)


def iterable_to_enum(iterable, name='CustomEnum'):
    return Enum(name, {str(kv): kv for kv in iterable})


def subdict(d: Mapping, keys=None):
    """Gets a sub-dict from a Mapping ``d``, extracting only those keys that are both in
    ``keys`` and ``d``.
    Note that the dict will be ordered as ``keys`` are, so can be used for reordering
    a Mapping.

    >>> subdict({'a': 1, 'b': 2, 'c': 3, 'd': 4}, keys=['b', 'a', 'd'])
    {'b': 2, 'a': 1, 'd': 4}
    """
    return {k: d[k] for k in (keys or ()) if k in d}


def _get_value_attr(d: dict, keys: Iterable, val_trans: Callable):
    """Return a copy of ``d`` where ``val_trans`` applied to the values of ``keys``.

    Meant to be used with ``functools.partial`` to fix ``val_trans``, and possibly
    ``transform_keys``.

    >>> _get_value_attr({'a': 1, 'b': 2, 'c': 3}, ['a', 'c'], lambda x: x * 10)
    {'a': 10, 'b': 2, 'c': 30}
    """
    return dict(d, **{k: val_trans(d[k]) for k in keys})


@double_up_as_factory
def inject_enum_annotations(func=None, *, extract_enum_value=True, **enum_list_for_arg):
    """

    :param func: function to wrap
    :param enum_list_for_arg: For every argument you want to enumify (i.e. annotate with
        an Enum), the list of enum values you want.
    :return: A wrapped func.

    >>> from inspect import signature
    >>> def foo(a, b: int, c='tat'):
    ...     return (a + b) * c
    >>> bar = inject_enum_annotations(foo, b=[1, 2], c='tit for tat'.split())
    >>> str(signature(bar))  # doctest: +SKIP
    '(a, b: front.util.b_enum, c: front.util.c_enum = 'tat')'
    >>> b_enum = Sig(bar).annotations['b']
    >>> list(b_enum)
    [<b_enum.1: 1>, <b_enum.2: 2>]
    >>>
    >>> c_enum = Sig(bar).annotations['c']
    >>> list(c_enum)
    [<c_enum.tit: 'tit'>, <c_enum.for: 'for'>, <c_enum.tat: 'tat'>]

    The wrapped bar now accepts enum objects (whose .values will be extracted and
    handed to foo for computation).

    >>> bar(a=2, b=b_enum['2'], c=c_enum.tit)
    'tittittittit'

    Again, note that the way we need to specify our ``b`` and ``c`` arguments are as
    `enum` types, not as the number `2` and string `'tit'`.

    In the following example, we ask for `extract_enum_value=False` so that we
    can still use normal inputs (no Enum objects, though ``b`` and ``c`` will
    still be annotated by the Enums.

    Also note in this example that you can use @ to decorate a function at
    definition time.


    >>> @inject_enum_annotations(
    ...     b=[1, 2], c='tit for tat'.split(), extract_enum_value=False
    ... )
    ... def foo(a, b: int, c='tat'):
    ...     return (a + b) * c
    >>>
    >>> foo(2, 2, 'tit')
    'tittittittit'

    """

    sig = Sig(func)
    with_enumed_sig = sig.ch_annotations(
        **{
            param: iterable_to_enum(enum_list, name=f'{param}_enum')
            for param, enum_list in enum_list_for_arg.items()
        }
    )

    if extract_enum_value:
        get_values_of_enums = partial(
            _get_value_attr, keys=list(enum_list_for_arg), val_trans=attrgetter('value')
        )

        dispatched_enums_func = wrap(
            func, ingress=Ingress(func, kwargs_trans=get_values_of_enums)
        )

        dispatched_enums_func.__signature__ = with_enumed_sig

        return with_enumed_sig(dispatched_enums_func)
    else:
        return with_enumed_sig(func)


from i2.signatures import empty
from i2.wrapper import Wrap
from typing import Union, Tuple, Callable, Any, Dict, Iterable

Annot = Any
AnnotForArgname = Union[Dict[str, Annot], Iterable[Tuple[str, Annot]]]
AnnotForType = Union[Dict[type, Annot], Iterable[Tuple[type, Annot]]]

# TODO: Compare to and possibly use i2.io_trans and/or i2.routing_forest.
# Note: A Parameter->annotation function would be more general
#   A Parameter->Parameter function even more general
def annotate_func_arguments(
    func: Callable,
    *,
    ignore_existing_annot: bool = False,
    annot_for_argname: AnnotForArgname = (),
    annot_for_dflt_type: AnnotForType = (),
    dflt_annot: Annot = empty,
):
    """Annotate

    :param func: The function whose args we want to annotate
    :param ignore_existing_annot: Set to True to ignore existing annots.
    :param annot_for_argname: Annotation for specific argnames
    :param annot_for_dflt_type: Annotation for specific types. Arg defaults will be
        compared (with ``isinstance(dflt_val, types)``) to types and the annotation
        (value) of the the first matching type (key) will be injected
    :param dflt_annot: Default annotation to use if no match found earlier.
        The default is `inspect.Parameter.empty`, which means "don't annotate".
        If you want all your params to be annotated no matter what, you might consider
        ``typing.Any``, or in the case of command line interfaces, ``str``.
    :return: A wrapped function with the desired signature changes, if any changes
        need to be made, or the same function untouched if not.

    >>> from inspect import signature
    >>> from functools import partial
    >>> from typing import Any
    >>>
    >>>
    >>> def foo(a, b, c, aa: int=1, bb: int=1.0, cc: int=None, aaa=1, bbb=1.0, ccc=None):
    ...     pass
    ...

    If nothing changes, you just get back the same function:

    >>> assert str(signature(annotate_func_arguments(foo))) == (
    ...     "(a, b, c, "
    ...     "aa: int = 1, bb: int = 1.0, cc: int = None, "
    ...     "aaa=1, bbb=1.0, ccc=None)"
    ... )

    In the following:

    - ``b: str`` through the argname rule, but ``bb`` (as well as ``aa`` and ``bb``)
    didn't change because ``ignore_existing_annot=False`` by default.

    - ``aaa: float`` (even though default is ``1``) and ``ccc: 'NoneAnnot'`` because of
    the ``annot_for_dflt_type`` rules.

    >>> annotator = partial(
    ...     annotate_func_arguments,
    ...     annot_for_argname = {'b': str, 'bb': str},
    ...     # don't confuse following with dict(int=float), which means {'int': float}
    ...     annot_for_dflt_type = {int: float, type(None): 'NoneAnnot'},
    ... )
    >>>
    >>> wrapped_func = annotator(foo)
    >>> assert str(signature(wrapped_func)) == (
    ... "(a, b: str, c, "
    ... "aa: int = 1, bb: int = 1.0, cc: int = None, "
    ... "aaa: float = 1, bbb=1.0, ccc: 'NoneAnnot' = None)"
    ... )

    See in the following what happens if we ask the default annotation to be ``Any`` and
    ``ignore_existing_annot=True``:

    >>> another_annotator = partial(
    ...     annotator,  # use the previous one, but...
    ...     dflt_annot=Any,  # and specify a default annotation
    ...     ignore_existing_annot=True  # now ignore any existing annotations
    ... )
    >>>
    >>> wrapped_func = another_annotator(foo)
    >>> assert str(signature(wrapped_func)) == (
    ... "(a: Any, b: str, c: Any, "
    ... "aa: float = 1, bb: str = 1.0, cc: 'NoneAnnot' = None, "
    ... "aaa: float = 1, bbb: Any = 1.0, ccc: 'NoneAnnot' = None)"
    ... )
    """
    annot_changes = dict(
        _annotate_func_arguments(
            func,
            annot_for_argname=annot_for_argname,
            annot_for_dflt_type=annot_for_dflt_type,
            dflt_annot=dflt_annot,
            ignore_existing_annot=ignore_existing_annot,
        )
    )
    if annot_changes:
        return Wrap(func, Ingress(Sig(func).modified(**annot_changes)))
    else:
        return func


def _annotate_func_arguments(
    func: Callable,
    *,
    annot_for_argname: AnnotForArgname = (),
    annot_for_dflt_type: AnnotForType = (),
    dflt_annot: Annot = empty,
    ignore_existing_annot=False,
):
    """Helper for annotate_func_arguments. Same inputs as the latter."""
    annot_for_argname = dict(annot_for_argname)
    annot_for_dflt_type = dict(annot_for_dflt_type)
    assert all(
        isinstance(t, str) and str.isidentifier(t) for t in annot_for_argname
    ), f'All keys should be identifier strings. Some were not: {annot_for_argname=}'
    assert all(
        isinstance(t, type) for t in annot_for_dflt_type
    ), f'All keys should be types. Some were not: {annot_for_dflt_type=}'
    handled_types = tuple(annot_for_dflt_type)

    for name, param in Sig(func).parameters.items():
        if ignore_existing_annot or param.annotation is empty:
            if name in annot_for_argname:
                yield name, {'annotation': annot_for_argname[name]}
            elif isinstance(default := param.default, handled_types):
                # NOTE: will yield the first one found
                for type_ in handled_types:
                    if isinstance(default, type_):
                        yield name, {'annotation': annot_for_dflt_type[type_]}
                        break
            else:
                if dflt_annot is not empty:
                    yield name, {'annotation': dflt_annot}


#
# f = annotate_func_arguments(
#     PCA,
#     annot_for_argname=dict(n_components=int, random_state=int),
#     annot_for_type={str: str}
# )
#
# Sig(f)


def get_value(obj, *args, **kwargs):
    return obj(*args, **kwargs) if isinstance(obj, Callable) else obj


def normalize_map(map: Map) -> Mapping:
    return get_value(map) or {}


def deep_merge(a: Mapping, b: Mapping):
    """Merges b into a"""

    result = dict(a)
    for key, value_b in b.items():
        value_a = a.get(key)
        if isinstance(value_a, Mapping) and isinstance(value_b, Mapping):
            result[key] = deep_merge(value_a, value_b)
        else:
            result[key] = value_b
    return result


def incremental_str_maker(str_format='{:03.f}'):
    """Make a function that will produce a (incrementally) new string at every call."""
    i = 0

    def mk_next_str():
        nonlocal i
        i += 1
        return str_format.format(i)

    return mk_next_str


unnamed_obj = incremental_str_maker(str_format='UnnamedObject{:03.0f}')


def obj_name(func):
    """The func.__name__ of a callable func, or makes and returns one if that fails.
    To make one, it calls unamed_func_name which produces incremental names to reduce the chances of clashing"""
    name = name_of_obj(func)
    if name is None or name == '<lambda>':
        return unnamed_obj()
    return name


def dflt_name_trans(obj):
    obj_str = obj if isinstance(obj, str) else obj_name(obj)
    return obj_str.replace('_', ' ').title()


def dflt_trans(objs):
    def trans(obj):
        trans_obj = copy(obj)
        trans_obj.__name__ = obj_name(obj)
        return trans_obj

    return list(map(trans, objs))



import os
from typing import Dict, Iterable
import pandas as pd


def add_ints(*nums_to_add: Iterable[int]):
    return sum(nums_to_add)


def add_floats(*nums_to_add: Iterable[float]):
    return sum(nums_to_add)


def add_strs(*strs_to_add: Iterable[str]):
    return ' '.join(strs_to_add)


def add_bools(*bools_to_add: Iterable[bool]):
    return sum(list(map(int, bools_to_add)))


def make_a_table_of_ints(col_name: str = 'integers', **keys_and_values: Dict[str, int]):
    return pd.Series(data=keys_and_values, name=col_name)


def make_a_table_of_floats(
    col_name: str = 'floats', **keys_and_values: Dict[str, float]
):
    return pd.Series(data=keys_and_values, name=col_name)


def make_a_table_of_strs(col_name: str = 'strings', **keys_and_values: Dict[str, str]):
    return pd.Series(data=keys_and_values, name=col_name)


def make_a_table_of_bools(
    col_name: str = 'booleans', **keys_and_values: Dict[str, bool]
):
    return pd.Series(data=keys_and_values, name=col_name)


funcs = [
    add_ints,
    add_floats,
    add_strs,
    add_bools,
    make_a_table_of_ints,
    make_a_table_of_floats,
    make_a_table_of_strs,
    make_a_table_of_bools,
]

if __name__ == '__main__':
    from streamlitfront.base import dispatch_funcs
    from streamlitfront.page_funcs import ArgsPageFunc

    print('file: {}'.format(os.path.realpath(__file__)))

    app = dispatch_funcs(funcs, configs={'page_factory': ArgsPageFunc})

    app()



# TODO: needs to be headless and minimum dependency

# from seleniumbase import BaseCase
# import time
#
#
#
# class ComponentsTest(BaseCase):
#     def test_basic(self):
#
#         # open the app and take a screenshot
#         self.open("http://localhost:8501")
#
#         time.sleep(5)
#         self.check_window(name="first_test", level=2)
#
#
#         self.assert_text("Identity")



# TODO: remove dependency to streamlitfront

# import pytest
# from streamlitfront import dispatch_funcs
# from front.scrap.pydantic_wrap import DFLT_CONFIGS, SimplePageFuncPydanticWrite
#
#
# def identity(x:int)->int:
# 	return x
#
# funcs = [identity]
#
# app = dispatch_funcs(funcs, configs=DFLT_CONFIGS)
# app()



"""Test combinations of objects of different modules"""


def foo(a, b):
    return a * b


def test_prepare_for_crude_dispatch_and_variations():
    from i2 import Sig

    # ------------ With prepare_for_crude_dispatch ----------

    from front.crude import prepare_for_crude_dispatch

    mall = {'a_store': {'one': 1, 'two': 2}, 'saves': {}}

    bar = prepare_for_crude_dispatch(
        foo,
        param_to_mall_map={'a': 'a_store'},
        mall=mall,
        output_store=mall['saves'],
        include_stores_attribute=True,
    )

    assert str(Sig(bar)) == "(a: str, b, save_name: str = '')"
    assert bar('two', 'mice', save_name='save_here') == 'micemice'

    assert bar.store_for_param == {'a': {'one': 1, 'two': 2}}
    assert bar.output_store == {'save_here': 'micemice'}

    # ------------ With prepare_for_dispatch (prepare_for_crude_dispatch + enum) ----------

    from front.base import prepare_for_dispatch

    mall = {'a_store': {'one': 1, 'two': 2}, 'saves': {}}

    bar = prepare_for_dispatch(
        foo,
        param_to_mall_map={'a': 'a_store'},
        mall=mall,
        output_store=mall['saves'],
        #     include_stores_attribute=True,
    )

    # Has the signature we expect (a had an enum annotation
    # and there's a new ``save_name`` argument

    assert str(Sig(bar)) == "(a: front.util.a_enum, b, save_name: str = '')"

    class FakeEnum:
        value = 'two'

    # Computes what we expect (if we feed it an object that has the input in a .value

    assert bar(FakeEnum, 'mice', save_name='save_here') == 'micemice'

    assert str(Sig(bar)) == "(a: front.util.a_enum, b, save_name: str = '')"

    # The output_store now contains the output under the specified ``save_name``
    assert mall['saves'] == {'save_here': 'micemice'}



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Iterable, List, Optional, TypedDict, Union
from i2 import Sig
from inspect import _empty
from front.types import FrontElementName
from front.util import deep_merge, get_value


@dataclass
class FrontElementBase(ABC):
    obj: Any = None
    name: FrontElementName = None

    def __post_init__(self):
        self.name = get_value(self.name, self.obj) or ''

    @abstractmethod
    def render(self):
        # raise NotImplementedError('This method needs to be implemented in subclasses.')
        pass


ELEMENT_KEY = '_front_element'
DEFAULT_INPUT_KEY = '_default'
FrontElementSpec = TypedDict('FrontElementSpec', {ELEMENT_KEY: FrontElementBase})


def mk_element_from_spec(spec: FrontElementSpec):
    _spec = dict(spec)
    try:
        factory = _spec.pop(ELEMENT_KEY)
    except KeyError:
        raise RuntimeError(
            f'Key "{ELEMENT_KEY}" is missing in the following element specification: \
                {spec}'
        )
    return factory(**_spec)


def mk_input_element_specs(obj, inputs, stored_value_getter):
    def mk_input_spec(p):
        input_spec = inputs_spec.get(p.name, {})
        annot = p.annotation if p.annotation != _empty else None
        param_type = annot or (type(p.default) if p.default != _empty else Any)
        type_spec = inputs_spec.get(param_type, {})
        input_spec = deep_merge(type_spec, input_spec)
        dflt_input_key = f'{obj.__name__}_{p.name}'
        input_key = input_spec.get('input_key', dflt_input_key)
        stored_value = stored_value_getter(input_key) if stored_value_getter else None
        init_value = (
            stored_value
            if stored_value is not None
            else (p.default if p.default != _empty else None)
        )
        return dict(input_spec, obj=p, input_key=input_key, init_value=init_value)

    inputs_spec = dict(inputs)
    default = inputs_spec.pop(DEFAULT_INPUT_KEY, {})
    inputs_spec = {k: deep_merge(default, v) for k, v in inputs_spec.items()}
    sig = Sig(obj)
    elements_spec = {p.name: mk_input_spec(p) for p in sig.params}
    return elements_spec


class FrontContainerBase(FrontElementBase):
    children: Iterable[FrontElementBase]

    def __init__(
        self, obj=None, name: FrontElementName = None, **kwargs: FrontElementSpec
    ):
        super().__init__(obj=obj, name=name)
        specs = [dict(dict(name=k, obj=obj), **v) for k, v in kwargs.items()]
        self._mk_children(specs)

    def _mk_children(self, specs):
        self.children = list(map(mk_element_from_spec, specs))

    def _render_children(self):
        for child in self.children:
            child.render()


@dataclass
class FrontComponentBase(FrontElementBase):
    pass


class TextSectionBase(FrontComponentBase):
    def __init__(
        self,
        content: str,
        kind: str = 'text',
        obj: Any = None,
        name: FrontElementName = None,
        **kwargs,
    ):
        super().__init__(obj, name)
        self.content = get_value(content, self.obj) or ''
        self.kind = get_value(kind, self.obj)
        self.kwargs = kwargs


@dataclass
class InputBase(FrontComponentBase):
    input_key: str = None
    init_value: Any = None


class OutputBase(FrontComponentBase):
    output: Any = None

    def render_output(self, output):
        self.output = output
        return self.render()


class ExecContainerBase(FrontContainerBase):
    def __init__(
        self,
        obj: Callable,
        inputs: dict,
        output: dict,
        name: FrontElementName = None,
        stored_value_getter: Callable[[str], Any] = None,
    ):
        element_specs = dict(
            mk_input_element_specs(obj, inputs, stored_value_getter), output=output
        )
        super().__init__(obj=obj, name=name, **element_specs)

    @property
    def input_components(self) -> Iterable[InputBase]:
        return [
            child
            for child in self.children
            if isinstance(child, InputBase)
            or isinstance(child, MultiSourceInputContainerBase)
        ]

    @property
    def output_component(self) -> OutputBase:
        return next(
            iter(child for child in self.children if isinstance(child, OutputBase))
        )


class MultiSourceInputContainerBase(FrontContainerBase):
    def __init__(
        self,
        obj=None,
        name: FrontElementName = None,
        input_key: str = None,
        init_value: Any = None,
        **kwargs: FrontElementSpec,
    ):
        # TODO: This is definitely not the right way to spread the input_key and
        # init_value to the child input components since a value can be compatible
        # with some compoenents and incompatible with others.
        # Just ignoring them for now.
        # kwargs = {
        #     k: dict(v, input_key=input_key, init_value=init_value)
        #     for k, v in kwargs.items()
        # }
        super().__init__(obj=obj, name=name, **kwargs)


@dataclass
class TextInputBase(InputBase):
    def __post_init__(self):
        super().__post_init__()
        self.init_value = str(self.init_value) if self.init_value is not None else ''


@dataclass
class NumberInputBase(InputBase):
    format: str = None


@dataclass
class IntInputBase(NumberInputBase):
    min_value: int = None
    max_value: int = None

    def __post_init__(self):
        super().__post_init__()
        self.init_value = int(self.init_value) if self.init_value is not None else 0


@dataclass
class FloatInputBase(NumberInputBase):
    min_value: float = None
    max_value: float = None
    step: float = None

    def __post_init__(self):
        super().__post_init__()
        self.init_value = float(self.init_value) if self.init_value is not None else 0.0


@dataclass
class FileUploaderBase(InputBase):
    type: Optional[Union[str, List[str]]] = None


@dataclass
class SelectBoxBase(InputBase):
    options: Iterable = None

    def __post_init__(self):
        super().__post_init__()
        self.options = self.options or []



from abc import ABC, abstractclassmethod
from typing import Any, Iterable
from front.elements.elements import (
    ELEMENT_KEY,
    FrontContainerBase,
)
from front.util import deep_merge


class ElementTreeMaker:
    def mk_tree(
        self, front_objs: Iterable[Any], rendering_spec: dict
    ) -> FrontContainerBase:
        self.front_objs = front_objs
        self.rendering_spec = dict(rendering_spec)
        root_factory = self._get_root_factory()
        obj_rendering_specs = {k: v for k, v in self._gen_obj_rendering_specs()}
        root: FrontContainerBase = root_factory(**obj_rendering_specs)
        return root

    def _get_root_factory(self):
        root_factory = self.rendering_spec.pop(ELEMENT_KEY)
        if not root_factory:
            raise RuntimeError(
                'No app element has been defined for this front application.'
            )
        return root_factory

    def _gen_obj_rendering_specs(self):
        for obj in self.front_objs:
            obj_name = obj.__name__
            type_rendering_spec = self._get_type_rendering_spec(
                self.rendering_spec, obj
            )
            obj_rendering_spec = self.rendering_spec.get(obj.__name__, {})
            obj_rendering_spec = deep_merge(type_rendering_spec, obj_rendering_spec)
            obj_rendering_spec['obj'] = obj
            yield (obj_name, obj_rendering_spec)

    def _get_type_rendering_spec(self, spec: dict, obj):
        obj_type = type(obj)
        obj_type_spec = spec.get(obj_type)
        if obj_type_spec:
            return obj_type_spec
        if getattr(obj_type, 'mro'):
            for t in obj_type.mro():
                type_spec = spec.get(t)
                if type_spec:
                    return type_spec
        for k, v in spec.items():
            if isinstance(obj, k):
                return v
        return {}



from abc import abstractclassmethod
import collections.abc
from inspect import isclass
from typing import Any, Mapping
from front.types import FrontSpec, Map
from front.util import deep_merge, dflt_name_trans, dflt_trans, normalize_map
from front.elements import *


APP_KEY = 'app'
OBJ_KEY = 'obj'
RENDERING_KEY = 'rendering'
NAME_KEY = 'name'

BASE_DFLT_CONVENTION = {
    APP_KEY: {'title': 'My Front Application'},
    OBJ_KEY: {'trans': dflt_trans},
    RENDERING_KEY: {
        collections.abc.Callable: {
            NAME_KEY: dflt_name_trans,
            'description': {NAME_KEY: 'Description', 'content': lambda o: o.__doc__,},
            'execution': {
                NAME_KEY: 'Execution',
                'inputs': {
                    float: {'format': '%.2f', 'step': 0.01,},
                    DEFAULT_INPUT_KEY: {NAME_KEY: lambda p: p.name},
                },
                'output': {NAME_KEY: 'Output',},
            },
        }
    },
}


class SpecMakerBase(ABC):
    def mk_spec(self, config: Map, convention: Map = None) -> FrontSpec:
        def get_inheritance_path(cls):
            path = []
            for cls_key in cls_keys:
                if cls != cls_key and issubclass(cls, cls_key):
                    i = 0
                    while i < len(path) and issubclass(path[i], cls_key):
                        i += 1
                    path.insert(i, cls_key)
            return path

        config = normalize_map(config)
        if convention is None:
            convention = deep_merge(BASE_DFLT_CONVENTION, self._dflt_convention)
        convention = normalize_map(convention)
        spec = deep_merge(convention, config)

        rendering_spec = spec.get('rendering', {})
        cls_keys = [k for k in rendering_spec if isclass(k)]
        inheritance_paths = {cls: get_inheritance_path(cls) for cls in cls_keys}
        for cls, path in inheritance_paths.items():
            cls_spec = rendering_spec[cls]
            for subcls in path:
                subcls_spec = rendering_spec[subcls]
                cls_spec = deep_merge(subcls_spec, cls_spec)
            rendering_spec[cls] = cls_spec

        return FrontSpec(
            obj_spec=spec.get(OBJ_KEY, {}),
            rendering_spec=rendering_spec,
            app_spec=spec.get(APP_KEY, {}),
        )

    @property
    @abstractclassmethod
    def _dflt_convention(cls) -> Mapping:
        pass



from typing import Any, Callable, Mapping, Union

Map = Union[None, Mapping, Callable[[], Mapping]]
Configuration = Mapping
Convention = Mapping
FrontElementName = Union[None, str, Callable[[Any], str]]


class FrontSpec:
    def __init__(self, obj_spec: dict, rendering_spec: dict, app_spec: dict):
        self.obj_spec = obj_spec
        self.rendering_spec = rendering_spec
        self.app_spec = app_spec


FrontApp = Callable



"""Wordle -- generalized"""

# The core functionality is


def wordle_feedback(guess, word):
    """The wordle score of a guessed word.

    >>> wordle_feedback('woopie', 'wordle')
    [2, 2, 1, 0, 0, 2]

    """
    if len(guess) != len(word):
        return (
            None  # instead of error, so it's less verbose to handle this case outside
        )
    return [
        # 0: no such letter, 1: wrong position, 2: right position
        (guess_letter in word_letter) + (guess_letter in word)
        for guess_letter, word_letter in zip(guess, word)
    ]


# Here's a CLI way the (dynamic) game can be played
# TODO: Better divide these steps into more abstract functions that would be able to
#  be reused
#  both on CLI and GUI for example.
# Note: And yes, we can do all kinds of things like count, or limit the guesses,
# but I'd like to keep it as simple as possible to show minimal code.
# TODO: Include a number (the number of letters) guessing game in this game instead of
#  telling the user how many words it has?
def guessing_game(
    word, prompt='Make a guess: ', char_for_score='.-*', stop_word='exit'
):
    char_for_score = {i: char for i, char in enumerate(char_for_score)}
    n = len(word)
    print(f'--> Type {stop_word} to exit the game')
    print(f'--> The word has {n} letters.')
    while (guess := input(prompt)) not in {word, stop_word}:
        feedback = wordle_feedback(guess, word)
        if feedback is None:
            print(f'--> The word must have {n} letters!')
        else:
            print(' ' * len(prompt) + f"{''.join(map(char_for_score.get, feedback))}")
    if guess != stop_word:
        print('--> Hurray!! The word was indeed {word}')
    print('--> Bye!')


# To play an actual game you have to not know the word to be guessed.
# So the word needs to be random -- but taken from a corpus of possible words.
# Here, we take the actual wordle words (big selection of 5-letter English words).
# TODO: Find source of words of other languages, and have the user chose their
#  language (and number of letters?)
from functools import lru_cache
import urllib.request

# The lru_cache ensures that you won't download the data twice (unless the cell/module
# is reloaded)
@lru_cache
def _get_url_text(url):
    with urllib.request.urlopen(url) as fp:
        return fp.read().decode('utf8')


def wordle_corpus():
    """A list of all wordle words (spoiler alert: past AND future)"""
    url = (
        'https://raw.githubusercontent.com/thorwhalen/content/master/tables/csv'
        '/wordle_words.txt'
    )
    return list(filter(None, _get_url_text(url).split('\n')))


# So here's an actual game (you can play it yourself for real since you don't know the
# word to guess!)
def start_a_game(corpus=None):
    import random

    corpus = corpus or wordle_corpus()
    word = random.choice(corpus)
    guessing_game(word)


if __name__ == '__main__':
    start_a_game()



"""Tools to dispatch dags

See below one of the dags that will often be used in this module's doctests:

>>> from meshed.makers import code_to_dag
>>> @code_to_dag
... def dag():
...     x = foo(a, b)
...     y = bar(x, greeting)
...     z = confuser(a, w=x)  # note the w=x to test non-trivial binding
>>> print(dag.dot_digraph_ascii())  # doctest: +SKIP

.. code-block::
     ┌──────────┐
  ┌▶ │ confuser │ ◀──    a
  │  └──────────┘
  │    │                │
  │    │                │
  │    ▼                ▼
  │                   ┌─────┐
  │       z           │ foo │ ◀──  b
  │                   └─────┘
  │                     │
  │                     │
  │                     ▼
  │
  └──────────────────    x

                        │
                        │
                        ▼
                      ┌─────┐
       greeting   ──▶ │ bar │
                      └─────┘
                        │
                        │
                        ▼

                         y



"""
from collections import defaultdict
from typing import Union, Iterable, Mapping, Any
from functools import partial
from itertools import chain

from operator import itemgetter
from enum import Enum
from dol import groupby
from front.crude import prepare_for_crude_dispatch

from meshed import DAG, FuncNode
from meshed.base import ch_func_node_attrs
from meshed.itools import parents, children


def simple_namer(name, *, prefix='', suffix=''):
    return f'{prefix}{name}{suffix}'


def crudify_func_nodes(
    var_nodes: Union[str, Iterable[str]],
    dag: DAG,
    var_node_name_to_store_name=partial(simple_namer, suffix='_store'),
    *,
    mall: Union[Mapping[str, Mapping[str, Any]], None] = None,
    include_stores_attribute: bool = False,
    save_name_param: str = 'save_name',
):
    """Crudifies the given ``var_nodes`` in the ``dag``.

    Crudifying a var node means crudifying it's ``FuncNode`` neighbors,
    i.e. telling the function that outputs to the ``VarNode`` (if any) to save it's
    output in a store and (additionally) return the key it saved it too instead of the
    value itself, and telling any consumers of the var node to use that key as it's
    argument instead, retrieving the value from said store.

    >>> from meshed import DAG, FuncNode
    >>> from inspect import signature
    >>> def foo(a, b):  return a + b
    >>> def bar(x, y):  return  x * y
    >>> dag = DAG([
    ...     FuncNode(foo, name='foo', out='foo_output'),
    ...     FuncNode(bar, bind={'y': 'foo_output'})
    ... ])

    Let's crudify ``'foo_output'``. We don't need to specify a mall, since
    ``crudify_func_nodes`` will make one for us.
    But in order to get access to it, to see what the function is doing, let's define
    a mall with a single store (a dictionary), named ``'foo_output_store'``
    (note that the map between ``var_node`` string name and
    store name is controlled by the ``var_node_name_to_store_name`` argument)

    >>> store = dict()
    >>> mall = {'foo_output_store': store}
    >>> new_dag = crudify_func_nodes(['foo_output'], dag, mall=mall)

    The ``new_dag`` is the same in structure, signature, and global behavior (you
    get the same outputs for the same inputs):

    >>> print(dag.synopsis_string())
    a,b -> foo -> foo_output
    foo_output,x -> bar_ -> bar
    >>> assert dag.synopsis_string() == new_dag.synopsis_string()
    >>> assert str(signature(dag)) == str(signature(new_dag)) == '(a, b, x)'
    >>> assert dag(2, 3, 4) == new_dag(2, 3, 4) == 20

    But let's have a closer look at the functions that ``dag`` and ``new_dag`` are
    using. The functions of the ``dag`` are the original functions we specified,
    behaving normally:

    >>> dag.func_nodes[0].func(2, 3)
    5
    >>> dag.func_nodes[1].func(4, 5)
    20

    But the first function of ``new_dag`` outputs ``'bar_last_output'`` instead of ``5``.

    >>> new_dag.func_nodes[0].func(2, 3)
    'bar_last_output'

    Where did the ``5`` go? In the mall!
    >>> mall
    {'foo_output_store': {'bar_last_output': 5}}

    So that ``5`` has been stored under the ``'bar_last_output'`` key.
    Further, the second function's second argument will no longer work with numbers,
    but with string keys, and use that same store to retrieve the value it needs for
    the underlying function:

    >>> new_dag.func_nodes[1].func(4, 'bar_last_output')
    20


    :param var_nodes: The ``VarNodes`` we want to crudify
    :param dag: The dag that contains these var_nodes
    :param var_node_name_to_store_name: The function to use to make a store for a given
        var_node name. If you have an explicit mapping ``m`` for this, just use ``m.get``
    :param mall: A ``mall`` (store of stores, i.e. mapping of mappings) whose keys are
        store names, and values are the actual stores.
    :param include_stores_attribute: Whether the crudified functions should have an
        attribute containing a pointer to the stores involved
    :param save_name_param: The name that the "save as" parameter should appear as.
    :return:
    """
    return DAG(
        list(
            _crudified_func_nodes(
                var_nodes,
                dag,
                var_node_name_to_store_name,
                mall=mall,
                include_stores_attribute=include_stores_attribute,
                save_name_param=save_name_param,
            )
        )
    )


class VarNodeRole(Enum):
    """(Var)Node roles.

    When a ``VarNode`` is used to source the arguments of a ``FuncNode``, it's playing
    a ``VarNodeRole.argument`` role.

    When a ``VarNode`` is used to store the return value of a ``FuncNode``, it's playing
    a ``VarNodeRole.return_value`` role.

    Most ``VarNode``s play both roles during a ``DAG`` computation.

    """

    argument = 'argument'
    return_value = 'return_value'


def _get_returned_by_func_node(var_node: str, dag: DAG):
    returned_by_func_node = parents(dag.graph, var_node)
    if len(returned_by_func_node) > 1:
        raise ValueError(
            f"This var_node had more than one parent. That is shouldn't be possible: "
            f'{var_node}'
        )
    return next(
        iter(returned_by_func_node), None
    )  # If None, it means it is NOT produced by a FuncNode, so is a root (input) node.


def _validate_is_func_node(node, var_node, relationship):
    if not isinstance(node, FuncNode):
        raise ValueError(f'{relationship} ({node}) of {var_node=} must be a FuncNode')


def _node_replacements_for_var_node_crudification(var_node: str, dag: DAG):
    """Helper function that generates the (node_id, (VarNodeRole, var_node)) instructions
    needed to crudify the input ``var_node`` in ``dag``.

    In the ``dag`` below, the ``var_node`` named ``x`` is the output of ``foo``
    and is used for the input of both ``bar`` (bound to the parameter of the same name)
    and ``confuser`` (bound to the ``w`` parameter).

    Therefore, as we "crudify" ``x`` we'll need to take care of all three cases:
    We'll need to have ``foo`` crudify it's output and both ``bar`` and ``confuser``
    crudify one of their params.

    >>> from meshed.makers import code_to_dag
    >>> @code_to_dag
    ... def dag():
    ...     x = foo(a, b)
    ...     y = bar(x, greeting)
    ...     z = confuser(a, w=x)  # note the w=x to test non-trivial binding
    >>> assert sorted(
    ...     _node_replacements_for_var_node_crudification('x', dag)
    ... ) == (
    ... [
    ...     ('bar', (VarNodeRole.argument, 'x')),
    ...     ('confuser', (VarNodeRole.argument, 'x')),
    ...     ('foo', (VarNodeRole.return_value, 'x')),
    ... ])

    """
    if not var_node in dag.var_nodes:
        raise ValueError(f"The {dag.name} dag doesn't have this var_node: {var_node}")

    returned_by_func_node = _get_returned_by_func_node(var_node, dag)
    if returned_by_func_node is not None:
        # If var_node is the output of a FuncNode
        # ... make sure it is.
        _validate_is_func_node(returned_by_func_node, var_node, 'Parent')
        # yield a (func_node_name, 'return_value', var_node) triple
        yield returned_by_func_node.name, (VarNodeRole.return_value, var_node)
    # if returned_by_func_node is None, skip the above: var_node is a "root" node

    # arg_of_func_nodes
    for arg_of_func_node in children(dag.graph, var_node):
        _validate_is_func_node(arg_of_func_node, var_node, 'Child')
        yield arg_of_func_node.name, (VarNodeRole.argument, var_node)


_no_more_elements = type('NoMoreElements', (), {})()


def _get_first_if_any_and_asserting_unique(
    iterable: Iterable,
    default=_no_more_elements,
    msg='Your iterator should have no more than one element',
):
    iterator = iter(iterable)
    first_element = next(iterator, default)
    assert next(iterator, _no_more_elements) == _no_more_elements, msg
    return first_element


def group_kvs_into_dict(kvs):
    return groupby(kvs, key=itemgetter(0), val=itemgetter(1))


# TODO: Skipping doctest because order not stable. Make it so
# TODO: Perhaps we should use FuncNode.name (id) instead of FuncNode itself as key of
#   dag.graph. We'd have less such problems then.
def _func_nodes_arg_and_return_names_to_crude(
    var_nodes: Union[str, Iterable[str]], dag: DAG,
):
    """Return a copy of a dag where ``var_nodes`` were crudified.

    >>> from meshed import DAG
    >>> from meshed.makers import code_to_dag
    >>> @code_to_dag
    ... def dag():
    ...     x = foo(a, b)
    ...     y = bar(x, greeting)
    ...     z = confuser(a, w=x)  # note the w=x to test non-trivial binding
    >>>
    >>> # Showing but skipping (because I can't get order to be stable (TODO: Make it so)
    >>> sorted(
    ... _func_nodes_arg_and_return_names_to_crude(['x', 'a'], dag)
    ... )  # doctest: +SKIP
    [
        (FuncNode(a,x -> confuser -> z), ('x', 'a'), None),
        (FuncNode(x,greeting -> bar -> y), ('x',), None),
        (FuncNode(a,b -> foo -> x), ('a',), 'x')
    ]

    """
    if isinstance(var_nodes, str):
        var_nodes = var_nodes.split()

    get_node_replacements = partial(
        _node_replacements_for_var_node_crudification, dag=dag
    )

    crude_modifications = group_kvs_into_dict(
        chain.from_iterable(map(get_node_replacements, var_nodes))
    )

    for func_node in dag.func_nodes:
        if modifications := crude_modifications.get(func_node.name, None):
            grouped_modifications = group_kvs_into_dict(modifications)
            argument_names = grouped_modifications.get(VarNodeRole.argument, [])
            return_name = _get_first_if_any_and_asserting_unique(
                iterable=grouped_modifications.get(VarNodeRole.return_value, []),
                default=None,
                msg=f"You shouldn't have more than one return_value in {func_node}",
            )
            yield func_node, tuple(argument_names), return_name
        else:
            yield func_node, (), None


def _mk_param_to_mall_map_from_for_var_nodes(
    var_nodes, bind, var_node_name_to_store_name
):
    """Generate (param, store_name) pairs.

    >>> dict(
    ...     _mk_param_to_mall_map_from_for_var_nodes(
    ...         var_nodes=['a', 'c'],
    ...         bind={'w': 'a', 'x': 'c', 'y': 'c', 'z': 'b'},
    ...         var_node_name_to_store_name=lambda name: f"{name}_store"
    ...     )
    ... )
    {'w': 'a_store', 'x': 'c_store', 'y': 'c_store'}
    """
    for param, var_node in bind.items():
        if var_node in var_nodes:
            yield param, var_node_name_to_store_name(var_node)


def _return_save_name(*, save_name) -> str:
    return save_name


def _empty_name_callback():
    raise RuntimeError(f'No save name was given')


from i2.wrapper import rm_params


def _crudified_func_nodes(
    var_nodes: Union[str, Iterable[str]],
    dag: DAG,
    var_node_name_to_store_name=partial(simple_namer, suffix='_store'),
    *,
    mall: Union[Mapping[str, Mapping[str, Any]], None] = None,
    include_stores_attribute: bool = False,
    save_name_param: str = 'save_name',
):
    if isinstance(var_nodes, str):
        var_nodes = var_nodes.split()
    # make a mall for all the var_names, giving them all empty dicts as stores if they are not there already.
    mall = mall or dict()
    mall = dict(
        {var_node_name_to_store_name(var_name): dict() for var_name in var_nodes},
        **mall,
    )
    rm_save_name = partial(rm_params, params_to_remove=[save_name_param])

    for (
        func_node,
        argument_names,
        return_name,
    ) in _func_nodes_arg_and_return_names_to_crude(var_nodes, dag):
        mk_param_mall_map = partial(
            _mk_param_to_mall_map_from_for_var_nodes,
            var_node_name_to_store_name=var_node_name_to_store_name,
        )
        if (argument_names, return_name) == ((), None):
            yield func_node
        else:
            param_to_mall_map = (
                dict(mk_param_mall_map(argument_names, func_node.bind)) or None
            )
            output_store = (
                var_node_name_to_store_name(return_name)
                if return_name is not None
                else None
            )

            crudified_func = prepare_for_crude_dispatch(
                func_node.func,
                param_to_mall_map=param_to_mall_map,
                output_store=output_store,
                empty_name_callback=None,
                auto_namer=lambda: f'{func_node.out}_last_output',
                output_trans=_return_save_name,
                mall=mall,
                include_stores_attribute=include_stores_attribute,
                save_name_param=save_name_param,
            )

            yield ch_func_node_attrs(func_node, func=rm_save_name(crudified_func))


# empty_name_callback: Callable[[], Any] = None,
# auto_namer: Callable[..., str] = None,
# output_trans: Callable[..., Any] = None,



# TODO: remove dependency to streamlitfront

# from streamlitfront import dispatch_funcs
# from front.scrap.pydantic_wrap import SimplePageFuncPydanticWithOutput
#
#
# def identity(x:int)->int:
# 	return x
#
# funcs = [identity]
#
# configs = {'page_factory': SimplePageFuncPydanticWithOutput}
#
#
# app = dispatch_funcs(funcs, configs=configs)
# app()



from distutils.command.config import config

from streamlitfront.base import mk_app


def foo(a: int = 1, b: int = 2, c=3):
    """This is foo. It computes something"""
    return (a * b) + c


def bar(x, greeting='hello'):
    """bar greets its input"""
    return f'{greeting} {x}'


def confuser(a: int, x: float = 3.14):
    return (a ** 2) * x


app = mk_app(
    [foo, bar, confuser],
    # config={
    #     'app': {
    #         'title': 'My app'
    #     }
    # }
)
app()



"""
Test prepare_for_crude_dispatch
"""

from front.crude import prepare_for_crude_dispatch
from functools import partial


# ---------------------------------------------------------------------------------------
# dispatchable function:
def apply_model(fitted_model, fvs, method='predict'):
    method_func = getattr(fitted_model, method)
    return method_func(list(fvs))


def learn_model(learner, fvs, method='fit'):
    method_func = getattr(learner, method)
    return method_func(list(fvs))


# a simple "learner" class to avoid importing from sklearn
class TinyMinMaxModel:
    def __init__(self):
        pass

    def fit(self, X, y=None):
        self.X_max = max(X)
        self.X_min = min(X)

        return self

    def predict(self, X):
        def min_max(x):
            return max(min(x, self.X_max), self.X_min)

        return [min_max(x) for x in X]


# The mall containing a store for each parameter of the apply_model and learn_model
# plus the model_results
mall = dict(
    learner=dict(TinyMinMaxModel=TinyMinMaxModel()),
    fvs=dict(train_fvs_1=[[1], [2], [3], [5], [4], [2], [1], [4], [3]],),
    fitted_model=dict(
        fitted_model_1=TinyMinMaxModel().fit(
            [[1], [2], [3], [5], [4], [2], [1], [4], [3]]
        )
    ),
    model_results=dict(),
)


def test_prepare_for_crude_dispatch():
    # Preparing the functions to be dispatched. The "dispatchable" versions will now
    # require a string for each parameter and rely on the mall above to call the
    # convert the string and call the original functions
    dispatchable_learn_model = prepare_for_crude_dispatch(
        learn_model,
        param_to_mall_map=dict(learner='learner', fvs='fvs'),
        mall=mall,
        output_store='fitted_model',
    )

    dispatchable_apply_model = prepare_for_crude_dispatch(
        apply_model,
        param_to_mall_map=dict(fitted_model='fitted_model', fvs='fvs'),
        mall=mall,
        output_store='model_results',
        save_name_param='save_name_for_apply_model',
    )

    # Here we set some defaults for both functions. This is useful in particular when
    # we want an UI to display defaults
    dispatchable_learn_model = partial(
        dispatchable_learn_model, learner='TinyMinMaxModel', fvs='train_fvs_1',
    )

    dispatchable_apply_model = partial(
        dispatchable_apply_model, fitted_model='fitted_model_1', fvs='train_fvs_1',
    )

    test_fvs = [[-100], [1], [2], [10]]

    # train a model with the dispatched function (with the defaults set above)
    tmm = dispatchable_learn_model()
    assert tmm.predict(test_fvs) == [[1], [1], [2], [5]]

    # run the dispatched function to apply a model. The output of which will saved in the
    # mall mall['model_results'], under the name 'apply_model_results'
    am = dispatchable_apply_model(save_name_for_apply_model='apply_model_results')
    assert am == [[1], [2], [3], [5], [4], [2], [1], [4], [3]]

    # checking that the store containing the results is created
    assert list(mall['model_results']) == ['apply_model_results']

    # checking the content of the results store
    assert mall['model_results']['apply_model_results'] == [
        [1],
        [2],
        [3],
        [5],
        [4],
        [2],
        [1],
        [4],
        [3],
    ]
